package journal.action;

import java.util.ArrayList;
import java.util.List;

import journal.reader.DeleteJournalEntry;
import journal.reader.DataJournalEntry;
import journal.reader.NoteJournalEntry;
import journal.reader.Options;
import journal.reader.TransactionJournalEntry;

public class CompoundAction implements Action 
{
	private List<Action> actions = new ArrayList<Action>();
	
	public void addAction(Action action) {
		actions.add(action);
	}
	
	@Override
	public void help() {
		for( Action a : actions ) 
			a.help();
	}
	
	@Override
	public void start() throws Exception {
		for( Action a : actions ) 
			a.start();
	}

	@Override
	public void setDefaultOptions(Options options) {
		// pass
	}
	
	@Override
	public void finish() throws Exception {
		for( Action a : actions ) 
			a.finish();
	}

	@Override
	public String[] parseArgs(String[] args) {
		for( Action a : actions ) {
			args = a.parseArgs(args);
		}
		return args;
	}

	@Override
	public void putValue(DataJournalEntry entry) throws Exception {
		for( Action a : actions ) 
			a.putValue(entry);
	}

	@Override
	public void replaceValue(DataJournalEntry entry) throws Exception {
		for( Action a : actions ) 
			a.replaceValue(entry);
	}

	@Override
	public void deleteValue(DataJournalEntry entry) throws Exception {
		for( Action a : actions ) 
			a.deleteValue(entry);
	}

	@Override
	public void verifyValue(DataJournalEntry entry) throws Exception {
		for( Action a : actions ) 
			a.verifyValue(entry);
	}

	@Override
	public void commitMarker(TransactionJournalEntry entry) throws Exception {
		for( Action a : actions ) 
			a.commitMarker(entry);
	}

	@Override
	public void flushMarker(TransactionJournalEntry entry) throws Exception {
		for( Action a : actions ) 
			a.flushMarker(entry);
	}

	@Override
	public void journalMarker(NoteJournalEntry entry) throws Exception {
		for( Action a : actions ) 
			a.journalMarker(entry);
	}
	
	@Override
	public void deleteMarker(DeleteJournalEntry entry) throws Exception {
		for( Action a : actions )
			a.deleteMarker(entry);
	}
}
